package biggests.gui;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.io.*;
import java.net.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;

import biggests.utils.*;
import de.schlichtherle.io.File;
import org.apache.batik.swing.svg.*;
import smadeira.biclustering.*;

/**
 * <p>Title: Function Analysis - Significance Graph Panel</p>
 *
 * <p>Description: Displays a significance graph of a bicluster.</p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonalves
 */

public class PanelAnalyzingFunctionSignificanceGraph extends JPanel {
    /**
     * PNG image type.
     */
    public static final String PNG = "png";

    /**
     * GIF image type.
     */
    public static final String GIF = "gif";

    /**
     * SVG image type.
     */
    public static final String SVG = "svg";

    /**
     * the application's main frame.
     */
    private BiggestsFrame mainFrame;

    /**
     * The layout of this panel.
     */
    private SpringLayout springLayout;

    /**
     * The path to image file (currently .png image).
     */
    private String pathToImage;

    /**
     * The path to svg file (.svg file).
     */
    private String pathToSvg;

    /**
     * A zoom slider to increase/decrease the size of the significance graph.
     */
    private JSlider zoomSlider;

    /**
     * The current zoom factor scale.
     */
    private float zoom;

    /**
     * The swing-based component which holds the svg graph.
     */
    private JSVGComponent svgGraph;

    /**
     * Flag which indicates when svg document has been fully parsed and
     * its tree has been built.
     */
    private boolean svgBuilded;

    /**
     * Creates a new function analysis - significance graph panel.
     *
     * @param frame the main <code>BiggestsFrame</code>
     */
    public PanelAnalyzingFunctionSignificanceGraph(BiggestsFrame frame) {
        this.mainFrame = frame;
        try {
            jbInit();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Graphic components initialization.
     *
     * @throws Exception if components cannot be initialized
     */
    private void jbInit() throws Exception {
        this.setBackground(Color.white);
        this.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 20));
        springLayout = new SpringLayout();
        this.setLayout(springLayout);
    }

    /**
     * Clears this panel. Removes the elements that are not needed anymore.
     */
    public void clear() {
        this.removeAll();
        svgGraph = null;
        this.springLayout = new SpringLayout();
        this.setLayout(springLayout);
    }

    /**
     * Paints this panel with a table of functions for the genes
     * in <code>bicluster</code>. The path to the given <code>bicluster</code>
     * datafile is retrieved from the <code>selectionPath</code>.
     *
     * @param bicluster the <code>Bicluster</code> from which the table
     *                  of functions will be constructed and shown
     * @param selectionPath the <code>TreePath</code> holding the necessary
     *                  data to find the path to the given
     *                  <code>bicluster</code> graph file
     */
    public void paint(Bicluster bicluster, TreePath selectionPath) {
        svgBuilded = false;
        Object[] path = selectionPath.getPath();

        /**
         * get bicluster's file path; it consists in:
         * the path to function results + "/" + dataset name + "_" +
         * bicluster group name + "/" + "graph-BICLUSTER_" + bicluster ID
         */
        String pathToBiclusterFiles = mainFrame.getPathToFunctionResults() +
                                      File.separator +
                                      path[1] +
                                      File.separator +
                                      ((DefaultMutableTreeNode) path[path.
                                       length - 2]).getUserObject() +
                                      File.separator +
                                      "graph-BICLUSTER_" +
                                      bicluster.getID();

        /**
         * define paths to .dot, .png and .svg files
         */
        String pathToBiclusterDotFile = pathToBiclusterFiles + ".dot";
        String pathToBiclusterImgFile = pathToBiclusterFiles + ".png";
        String pathToBiclusterSvgFile = pathToBiclusterFiles + ".svg";

        /**
         * create zoom slider
         */
        zoomSlider = new JSlider(0, 100, 50);
        zoomSlider.setMajorTickSpacing(50);
        zoomSlider.setMinorTickSpacing(50);
        zoomSlider.setPaintLabels(true);
        zoomSlider.setBackground(Color.white);
        zoomSlider.addChangeListener(
            new
            PanelAnalyzingFunctionSignificanceGraph_zoomSlider_changeAdapter(this));
        zoom = (float) 0.5;

        File biclusterDotFile = new File(pathToBiclusterDotFile);

        if (!biclusterDotFile.exists() || !biclusterDotFile.isFile()) {
            JOptionPane.showMessageDialog(this, "Unable to find .dot file " +
                                          "to process: " + pathToBiclusterDotFile,
                                          "Unable to find .dot file",
                                          JOptionPane.ERROR_MESSAGE);
            return;
        }

        /**
         * convert .dot file to .png image
         */
        File biclusterImgFile = new File(pathToBiclusterImgFile);
        this.pathToImage = biclusterImgFile.getAbsolutePath();
        try {
            this.convertDOTToImage(biclusterDotFile,
                                   biclusterImgFile, this.PNG);
        } catch (IOException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(this, "Unable to convert .dot " +
                                          "file to image: " + ex.getMessage() +
                                          ".",
                                          "Unable to convert .dot file",
                                          JOptionPane.ERROR_MESSAGE);
            ex.printStackTrace();
            return;
        } catch (InterruptedException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(this, "Unable to convert .dot " +
                                          "file to image.",
                                          "Unable to convert .dot file",
                                          JOptionPane.ERROR_MESSAGE);
            ex.printStackTrace();
            return;
        }

        /**
         * convert .dot file to .svg document
         */
        File biclusterSvgFile = new File(pathToBiclusterSvgFile);
        this.pathToSvg = biclusterSvgFile.getAbsolutePath();
        try {
            this.convertDOTToImage(biclusterDotFile,
                                   biclusterSvgFile, this.SVG);
        } catch (IOException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(this, "Unable to convert .dot " +
                                          "file to .svg: " + ex.getMessage() +
                                          ".",
                                          "Unable to convert .dot file",
                                          JOptionPane.ERROR_MESSAGE);
            return;
        } catch (InterruptedException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(this, "Unable to convert .dot " +
                                          "file to .svg",
                                          "Unable to convert .dot file",
                                          JOptionPane.ERROR_MESSAGE);
            return;
        }

        /**
         * load svg graph document
         */
        svgGraph = new JSVGComponent();
        try {
            svgGraph.loadSVGDocument(biclusterSvgFile.toURI().toURL().toString());
        } catch (MalformedURLException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(this, "Unable to display graph: " +
                                          "\ninvalid svg file after dot conversion.",
                                          "Unable to display graph",
                                          JOptionPane.ERROR_MESSAGE);
            return;
        }
        final PanelAnalyzingFunctionSignificanceGraph thisPanel = this;

        svgGraph.addGVTTreeBuilderListener(new GVTTreeBuilderAdapter() {
            /**
             * Handles the event triggered when the building of the GVT
             * tree for the current SVG document is completed.
             *
             * @param evt <code>GVTTreeBuilderEvent</code> fired when
             *            the building of the gvt tree is completed
             */
            public void gvtBuildCompleted(GVTTreeBuilderEvent evt) {
                svgBuilded = true;
                AffineTransform at = new AffineTransform();
                at.setToScale(0.5, 0.5);
                svgGraph.setRenderingTransform(at);
                Dimension2D size = svgGraph.getSVGDocumentSize();
                Dimension newPanelSize = new Dimension((int) (size.getWidth() *
                        zoom + 80), (int) (size.getHeight() * zoom) + 100);

                thisPanel.setSize(newPanelSize);
                thisPanel.setPreferredSize(newPanelSize);
                thisPanel.setMaximumSize(newPanelSize);
                thisPanel.setMinimumSize(newPanelSize);
                mainFrame.setVisible(true);
            }
        });

        /**
         * display graphic components
         */
        JButton saveGraph = new JButton("Save Graph");
        saveGraph.addActionListener(new
                                   PanelAnalyzingFunctionSignificanceGraph_saveGraph_actionAdapter(this));
        this.add(saveGraph);
        this.add(zoomSlider);
        springLayout.putConstraint(SpringLayout.WEST, zoomSlider, 50,
                                   SpringLayout.EAST, saveGraph);
        springLayout.putConstraint(SpringLayout.NORTH, zoomSlider, 0,
                                   SpringLayout.NORTH, saveGraph);
        this.add(svgGraph);
        springLayout.putConstraint(SpringLayout.WEST, svgGraph, 20,
                                   SpringLayout.WEST, saveGraph);
        springLayout.putConstraint(SpringLayout.NORTH, svgGraph, 40,
                                   SpringLayout.SOUTH, saveGraph);
        mainFrame.setVisible(true);
    }

    /**
     * Converts a .dot file to an image file (png, gif, svg).
     *
     * @param pathToDotFile the <code>String</code> path to dot file
     * @param pathToImgDirectory the <code>String</code> path to destination directory
     * @param imgFileExtension the <code>String</code> image file extension
     * @return the <code>String</code> path to the new image file
     * @throws InterruptedException if an error occured while waiting for the conversion of the .dot file
     * @throws IOException if files do not exist or an error occurs in the conversion process
     * @throws Exception if it was not possible to convert the .dot file
     */
    public String convertDOTtoImage(String pathToDotFile,
                                    String pathToImgDirectory,
                                    String imgFileExtension) throws
            InterruptedException, IOException, Exception {
        if (!(imgFileExtension.toLowerCase().equals("png")) &&
            !(imgFileExtension.toLowerCase().equals("gif")) &&
            !(imgFileExtension.toLowerCase().equals("svg"))) {
            throw new Exception("Image file extension " + imgFileExtension +
                                " is not supported.");
        }

        if (!pathToDotFile.toLowerCase().endsWith(".dot")) {
            throw new Exception(
                    "The path to dot file does not contain a dot file.");
        }

        File dotFile = new File(pathToDotFile);
        if (!dotFile.exists() || !dotFile.isFile()) {
            throw new Exception(
                    "The path to dot file does not contain a valid dot file.");
        }

        File imageDirectory = new File(pathToImgDirectory);
        if (!imageDirectory.exists() || !imageDirectory.isDirectory()) {
            throw new Exception("The path to image directory is not valid.");
        }

        String imgFilename = dotFile.getName().split(".")[0] + "." +
                             imgFileExtension;
        File imageFile = new File(imageDirectory.getAbsolutePath() +
                                  File.separator + imgFilename);

        Runtime runtime = Runtime.getRuntime();
        String[] commands = {mainFrame.getPathToDOT(),
                            "-T" + imgFileExtension,
                            dotFile.getAbsolutePath(),
                            "-o" + imageFile.getAbsolutePath()};
        try {
            Process process = runtime.exec(commands);
            process.waitFor();
        } catch (IOException ex) {
            throw new IOException("couldn't find dot binary file to " +
                                  "convert dot files;\ndot should be " +
                                  "located in " + mainFrame.getPathToDOT() +
                                  "\nYou won't be able to see Significance Graph");
        }

        return imageFile.getAbsolutePath();
    }

    /**
     * Converts a .dot file to an image file (png, gif, svg).
     *
     * @param dotFile the .dot <code>File</code>
     * @param imgFile the image <code>File</code>
     * @param extension the <code>String</code> image file extension
     * @throws IOException if files do not exist or an error occurs in the conversion process
     * @throws InterruptedException if an error occured while waiting for the conversion of the .dot file
     */
    private void convertDOTToImage(File dotFile, File imgFile, String extension) throws
            IOException, InterruptedException {
        Runtime runtime = Runtime.getRuntime();
        String[] commands = {mainFrame.getPathToDOT(),
                            "-T" + extension,
                            dotFile.getAbsolutePath(),
                            "-o" + imgFile.getAbsolutePath()};

        try {
            Process process = runtime.exec(commands);
            process.waitFor();
        } catch (IOException ex) {
            throw new IOException("couldn't find dot binary file to " +
                                  "convert dot files;\ndot should be " +
                                  "located in " + mainFrame.getPathToDOT() +
                                  "\nYou won't be able to see Significance Graph");
        }
    }

    /**
     * Handles the user click on the save graph button.
     * It allows the saving of the current significance graph in png or
     * svg format.
     *
     * @param e the <code>ActionEvent</code> fired by the click of the
     *          user in save graph button
     */
    public void saveGraph_actionPerformed(ActionEvent e) {
        JFileChooser pathFileChooser = new JFileChooser();
        try {
            /**
             * initialize the file chooser with the intended file filters
             * (one for png files, and another for svg files)
             */
            pathFileChooser.addChoosableFileFilter(new FileFilterDirectory(
                    mainFrame.getPathsToFolders(), ".png"));
            pathFileChooser.addChoosableFileFilter(new FileFilterDirectory(
                    mainFrame.getPathsToFolders(), ".svg"));
        } catch (Exception ex) {
            ex.printStackTrace();
        }

        /**
         * turn off accept all file filter
         */
        pathFileChooser.setAcceptAllFileFilterUsed(false);

        /**
         * show the save dialog to the user
         */
        int option = pathFileChooser.showSaveDialog(this.mainFrame);

        if (option == JFileChooser.APPROVE_OPTION) {
            /**
             * user hit approve option; get the specified path for the file
             */
            String newPath = pathFileChooser.getSelectedFile().getPath();
            newPath = ((FileFilterDirectory) pathFileChooser.getFileFilter()).
                      checkAndAddFileExtension(newPath);

            File newFile = new File(newPath);
            if (newFile.exists()) {
                if (newFile.isDirectory()) {
                    JOptionPane.showMessageDialog(this.mainFrame,
                                                  "Unable to save .png image." +
                                                  "\nYou didn't specify a filename.",
                                                  "Filename missing",
                                                  JOptionPane.ERROR_MESSAGE);
                    return;
                }
                if (newFile.isFile()) {
                    int opt = JOptionPane.showOptionDialog(this.mainFrame,
                            "The specified file already exists." +
                            "\nOverwrite?",
                            "Overwrite?",
                            JOptionPane.YES_NO_OPTION,
                            JOptionPane.QUESTION_MESSAGE,
                            null, null, JOptionPane.YES_OPTION);
                    if (opt == JOptionPane.NO_OPTION) {
                        /**
                         * the user does not want to overwrite the file
                         */
                        return;
                    }
                }
            }

            String extension = ((FileFilterDirectory) pathFileChooser.
                                getFileFilter()).getValidFileExtension();
            File currentFile = null;
            if (extension.equals(".png")) {
                currentFile = new File(this.pathToImage);
            } else {
                currentFile = new File(this.pathToSvg);
            }
            if (currentFile == null) {
                JOptionPane.showMessageDialog(this.mainFrame,
                                              "Unable to save " + extension +
                                              " image.",
                                              "Unable to save " + extension,
                                              JOptionPane.ERROR_MESSAGE);
                return;
            }
            try {
                /**
                 * save the specified file (png or svg file) in the correct location
                 */
                BiggestsIO.copy(currentFile, newFile);
            } catch (IOException ex) {
                ex.printStackTrace();
                JOptionPane.showMessageDialog(this.mainFrame,
                                              "Unable to save " + extension +
                                              " image.",
                                              "Unable to save " + extension,
                                              JOptionPane.ERROR_MESSAGE);
                return;
            }
        }
    }

    /**
     * Handles changes on the zoom slider.
     *
     * @param e <code>ChangeEvent</code> fired by the user manipulation of the zoom slider
     */
    public void zoomSlider_stateChanged(ChangeEvent e) {
        if (!svgBuilded) {
            return;
        }

        /**
         * get the zoom scale factor selected on the zoom slider
         */
        zoom = (float) (zoomSlider.getValue() * 1.0 / 100);

        /**
         * get svg original document size
         */
        Dimension2D size = svgGraph.getSVGDocumentSize();

        /**
         * apply the zoom factor scale to the svg graph and panel's size
         */
        Dimension newGraphSize = new Dimension((int) (size.getWidth() * zoom),
                                               (int) (size.getHeight() * zoom));
        Dimension newSize = new Dimension((int) (size.getWidth() * zoom) + 80,
                                          (int) (size.getHeight() * zoom) + 100);

        /**
         * create a new rendering transform, with the intended zoom factor
         * scale and apply it to the svg graph
         */
        AffineTransform at = new AffineTransform();
        at.setToScale(zoom, zoom);
        svgGraph.setRenderingTransform(at);

        /**
         * set svg graph new size
         */
        svgGraph.setSize(newGraphSize);
        svgGraph.setPreferredSize(newGraphSize);
        svgGraph.setMinimumSize(newGraphSize);
        svgGraph.setMaximumSize(newGraphSize);

        /**
         * set this panel's new size
         */
        this.setSize(newSize);
        this.setPreferredSize(newSize);
        this.setMaximumSize(newSize);
        this.setMinimumSize(newSize);
        this.repaint();
    }

    /**
     * Returns the application's main frame.
     *
     * @return the main <code>BiggestsFrame</code>
     */
    public BiggestsFrame getMainFrame() {
        return mainFrame;
    }

    /**
     * Returns the path to image.
     *
     * @return the path <code>String</code> to image
     */
    public String getPathToImage() {
        return pathToImage;
    }

    /**
     * Returns the path to svg.
     *
     * @return the path <code>String</code> to svg
     */
    public String getPathToSvg() {
        return pathToSvg;
    }

    /**
     * Returns the zoom slider.
     *
     * @return the zoom <code>JSlider</code>
     */
    public JSlider getZoomSlider() {
        return zoomSlider;
    }

    /**
     * Returns the current zoom scale factor.
     *
     * @return the <code>float</code> zoom scale factor
     */
    public float getZoom() {
        return zoom;
    }

    /**
     * Returns the swing-based component which holds the svg graph.
     *
     * @return the swing-based <code>JSVGComponent</code> which holds the svg graph
     */
    public JSVGComponent getSvgGraph() {
        return svgGraph;
    }

    /**
     * Checks if the gvt tree for the svg document is already built.
     *
     * @return <code>boolean</code> that indicates if gvt tree for the svg document is already built
     */
    public boolean isSvgBuilded() {
        return svgBuilded;
    }
}


/**
 * <p>Title: </p>
 *
 * <p>Description: </p>
 */
class PanelAnalyzingFunctionSignificanceGraph_zoomSlider_changeAdapter implements
        ChangeListener {
    private PanelAnalyzingFunctionSignificanceGraph adaptee;
    PanelAnalyzingFunctionSignificanceGraph_zoomSlider_changeAdapter(
            PanelAnalyzingFunctionSignificanceGraph adaptee) {
        this.adaptee = adaptee;
    }

    /**
     * Handles the events of changing in the state of certain components.
     *
     * @param e the <code>ChangeEvent</code> fired by a user's selection
     */
    public void stateChanged(ChangeEvent e) {
        adaptee.zoomSlider_stateChanged(e);
    }
}


/**
 * <p>Title: </p>
 *
 * <p>Description: </p>
 */
class PanelAnalyzingFunctionSignificanceGraph_saveGraph_actionAdapter implements
        ActionListener {
    private PanelAnalyzingFunctionSignificanceGraph adaptee;
    PanelAnalyzingFunctionSignificanceGraph_saveGraph_actionAdapter(
            PanelAnalyzingFunctionSignificanceGraph adaptee) {
        this.adaptee = adaptee;
    }

    /**
     * Handles events fired by mouse click by the user.
     *
     * @param e the <code>MouseEvent</code> triggered by a user's click
     */
    public void actionPerformed(ActionEvent e) {
        adaptee.saveGraph_actionPerformed(e);
    }
}
