package biggests.gui;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.tree.*;

import biggests.tables.*;
import biggests.tree.*;
import biggests.utils.*;
import de.schlichtherle.io.*;
import smadeira.biclustering.*;
import smadeira.ontologizer.*;
import smadeira.utils.BiologicalRelevanceSummary;
import java.io.IOException;
import ontologizer.go.OBOParserException;

/**
 * <p>Title: Panel for Analyzing Tables of Functions</p>
 *
 * <p>Description: Displays a table of functions of a bicluster
 *                 or group of biclusters.</p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonalves
 */

public class PanelAnalyzingFunctionTable extends JPanel {
    /**
     * This panel's layout.
     */
    private BoxLayout boxLayout;

    /**
     * The application's main frame.
     */
    private BiggestsFrame mainFrame;

    /**
     * The function analysis panel.
     */
    private PanelFunctionAnalysis functionAnalysisPanel;

    /**
     * The selection path.
     */
    private TreePath selectionPath;

    /**
     * The object.
     */
    private NodeObjectInterface nodeObject;

    /**
     * All computed GO frontends.
     */
    private HashMap<String, GOFrontEnd> goFrontEnds;

    /**
     * Creates a new <code>PanelAnalyzingFunctionTable</code>.
     *
     * @param frame the main <code>BiggestsFrame</code>
     */
    public PanelAnalyzingFunctionTable(BiggestsFrame frame) {
        this.mainFrame = frame;
        goFrontEnds = new HashMap<String, GOFrontEnd>();
        try {
            jbInit();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Component initialization.
     *
     * @throws Exception if the components could not be initialized
     */
    private void jbInit() throws Exception {
        this.setPreferredSize(new Dimension(500, 500));
        boxLayout = new BoxLayout(this, BoxLayout.Y_AXIS);
        this.setLayout(this.boxLayout);
        this.setBorder(BorderFactory.createEmptyBorder(50, 50, 50, 50));
        functionAnalysisPanel = new PanelFunctionAnalysis(mainFrame);
    }

    public void initializeGOFrontEnds() {
        goFrontEnds = new HashMap<String, GOFrontEnd>();
    }

    /**
     * Returns function analysis panel.
     *
     * @return PanelFunctionAnalysis the panel that holds function
     *         analysis options
     */
    public PanelFunctionAnalysis getFunctionAnalysisPanel() {
        return functionAnalysisPanel;
    }

    /**
     * Returns the tree selection path that corresponds to the bicluster
     * or bicluster group that is being processed.
     *
     * @return TreePath the selection path that corresponds to the bicluster
     *                  or bicluster group being processed in this panel.
     */
    public TreePath getSelectionPath() {
        return selectionPath;
    }

    /**
     * Returns the bicluster group that is being processed.<br/>
     * Null if the object being processed is not a bicluster group
     * (it is a single bicluster instead).
     *
     * @return Biclustering the bicluster group that is being processed
     */
    public NodeObjectInterface getNodeObject() {
        return nodeObject;
    }

    /**
     * Clears this panel's components.
     */
    public void clear() {
        this.removeAll();
        this.functionAnalysisPanel.clear();
        nodeObject = null;
        selectionPath = null;
    }

    /**
     * Paints the function analysis panel.
     */
    private void paintFunctionAnalysisPanel() {
        functionAnalysisPanel.paint(selectionPath);

        JButton applyButton = new JButton("Apply");
        applyButton.setPreferredSize(new Dimension(400, 50));
        SpringLayout sLPanelApplyFunctionAnalysis = new SpringLayout();
        JPanel panelApplyFunctionAnalysis = new JPanel(
                sLPanelApplyFunctionAnalysis);

        panelApplyFunctionAnalysis.add(functionAnalysisPanel);
        panelApplyFunctionAnalysis.add(applyButton);

        sLPanelApplyFunctionAnalysis.putConstraint(
                SpringLayout.WEST, applyButton, 0,
                SpringLayout.WEST, functionAnalysisPanel);
        sLPanelApplyFunctionAnalysis.putConstraint(
                SpringLayout.NORTH, applyButton, 20,
                SpringLayout.SOUTH, functionAnalysisPanel);

        this.add(panelApplyFunctionAnalysis);

        applyButton.addActionListener(new
                                     PanelAnalyzingFunctionTable_applyButton_actionAdapter(this));
    }

    /**
     * Paints a table of functions with data retrieved from the given
     * <code>expressionMatrix</code>.
     *
     * @param expressionMatrix the <code>IMatrix</code> with the
     *        function information
     * @param selectionPath <code>TreePath</code> selected in the
     *        dataset tree, whose node contains the given
     *        <code>expressionMatrix</code>
     */
    public void paint(IMatrix expressionMatrix, TreePath selectionPath) {
        this.nodeObject = expressionMatrix;
        this.selectionPath = selectionPath;
        GeneSetResult goProperties =
            expressionMatrix.getGeneOntologyProperties();
        if (goProperties == null) {

//        GOFrontEnd go = this.getGOFrontEndForSelectionPath(selectionPath);

//        if (goProperties == null) {

/*            String organism = null;
            if (expressionMatrix instanceof AbstractExpressionMatrix) {
                organism = ((AbstractExpressionMatrix) expressionMatrix).
                           getOrganism();
            } else {
                organism = ((DiscretizedExpressionMatrix) expressionMatrix).
                           getOriginalExpressionMatrix().getOrganism();
            }

            String pathToGeneOntologyFile = mainFrame.getPathToGO_Files() +
                                            File.separator +
                                            "gene_ontology.obo";
            String pathToGeneAssociationFile =
                    mainFrame.getPathToGO_Files() +
                    File.separator +
                    "gene_association" +
                    PanelFunctionAnalysis.
                    findAnnotationFileExtensionForOrganism(organism);
 */

try {
    ThreadFunctionAnalysis functionThread =
                        new ThreadFunctionAnalysis(
        mainFrame,
        expressionMatrix,
        selectionPath,
        /*                        pathToGeneOntologyFile,
                                                          pathToGeneAssociationFile,
         */
                this.getGOFrontEndForSelectionPath(selectionPath),
new File(mainFrame.getPathToFunctionResults() +
                                 File.separator +
                                 selectionPath.getPath()[1] +
                                 File.separator + "Matrices"),
                        ((DefaultMutableTreeNode) selectionPath.
                         getLastPathComponent()).
                        getUserObject().toString(),
                        (float) 0.05);
            } catch (Exception ex) {
                ex.printStackTrace();
                mainFrame.getTabbedPaneAnalyzing().setEnabledAt(2, false);
                return;
            }
        }

        goProperties = expressionMatrix.getGeneOntologyProperties();
        if (goProperties == null) {
            mainFrame.getTabbedPaneAnalyzing().setEnabledAt(2, false);
            return;
        }
/*        GOFrontEnd go = this.getGOFrontEndForSelectionPath(selectionPath);
        if (go == null) {
            go = this.calculateGOFrontEnd(expressionMatrix);
            this.putGOFrontEnd(selectionPath, go);
        }
 */

        try {
            TableFunctionIndividual functionTable = new
                    TableFunctionIndividual(expressionMatrix,
            this.getGOFrontEndForSelectionPath(selectionPath));
            JScrollPane scrollTable = new JScrollPane(functionTable);
            scrollTable.setBorder(BorderFactory.createEmptyBorder(20, 0, 20, 0));
            scrollTable.getVerticalScrollBar().
                setUnitIncrement(mainFrame.SCROLL_BAR_INCREMENT/2);

            this.add(scrollTable);
        } catch (InvalidNodeObjectTypeException ex) {
            ex.printStackTrace();
            mainFrame.getTabbedPaneAnalyzing().setEnabledAt(2, false);
            return;
        } catch (Exception ex) {
            // gene set result is empty
            mainFrame.getTabbedPaneAnalyzing().setEnabledAt(2, false);
            return;
        }
        mainFrame.getTabbedPaneAnalyzing().setEnabledAt(2, true);
        mainFrame.getTabbedPaneAnalyzingFunction().setEnabledAt(0, true);
        mainFrame.getTabbedPaneAnalyzingFunction().setEnabledAt(1, false);
    }

    /**
     * Paints a function table with the GO terms annotated for all the
     * genes of the experiment.
     *
     * @param goProperties <code>GeneSetResult</code> with GO terms
     * @param go the <code>GOFrontEnd</code>
     * @param genesNames <code>String[]</code> the set of genes' names
     * @param pvalue <code>float</code> the pvalue threshold
     */
    public void paint(GeneSetResult goProperties, GOFrontEnd go,
                      String[] genesNames, float pvalue) {
        try {
            TableFunctionIndividual biclusterFunctionTable = new
                    TableFunctionIndividual(goProperties, go, genesNames,
                                            pvalue);
            JScrollPane scrollTable = new JScrollPane(biclusterFunctionTable);
            scrollTable.setBorder(BorderFactory.createEmptyBorder(20, 0, 20, 0));
            scrollTable.getVerticalScrollBar().
                setUnitIncrement(mainFrame.SCROLL_BAR_INCREMENT/2);

            this.add(scrollTable);
        } catch (InvalidNodeObjectTypeException ex) {
            ex.printStackTrace();
            return;
        } catch (Exception ex) {
            //gene set result is empty; disable function tab
            mainFrame.getTabbedPaneAnalyzing().setEnabledAt(2, false);
            return;
        }
    }

    /**
     * Paints the functions table in this panel, with data retrieved from
     * <code>bicluster</code> object.
     *
     * @param bicluster the <code>Bicluster</code> object which contains the
     *                  function analysis to present in the functions table
     * @param goProperties the <code>GeneSetResult</code> for the given
     *                     <code>bicluster</code>
     * @param pvalue the p-value threshold for significance
     * @param selectionPath the <code>TreePath</code> selected in the
     *                  dataset tree
     */
    public GeneSetResult paint(Bicluster bicluster,
                      GeneSetResult goProperties,
                      float pvalue,
                      TreePath selectionPath) {
        this.selectionPath = selectionPath;
        this.nodeObject = bicluster;
        GOFrontEnd go = null;
        if (goProperties == null) {
            if ((go = this.getGOFrontEndForSelectionPath(selectionPath)) == null) {
                go = calculateGOFrontEnd(bicluster);
                this.putGOFrontEnd(selectionPath, go);
                if (go == null) {
                    this.paintFunctionAnalysisPanel();
                    mainFrame.getTabbedPaneAnalyzingFunction().setEnabledAt(1, false);
                    return goProperties;
                }
            }
            goProperties = bicluster.computeGeneOntologyTerms(go, false);
        }


        JLabel genesInPopulation = new JLabel("Number of Genes in Population: " +
                                              goProperties.
                                              getPopulationGeneCount());
        JLabel genesInBicluster = new JLabel("Number of Genes in Bicluster: " +
                                             bicluster.getNumberOfGenes());

        JLabel pvalueInfo = new JLabel("Values calculated using p-value = " +
                                       pvalue);
        JButton recalculate = new JButton("Recalculate");
        recalculate.addActionListener(
                new
                PanelAnalyzingFunctionTable_recalculate_actionAdapter(this));
        pvalueInfo.setPreferredSize(new Dimension(
                (int) pvalueInfo.getPreferredSize().getWidth(),
                (int) recalculate.getPreferredSize().getHeight()));

        SpringLayout springLayoutPanelPValueRecalc = new SpringLayout();
        JPanel panelPValueRecalc = new JPanel(springLayoutPanelPValueRecalc);
        panelPValueRecalc.setPreferredSize(new Dimension(400, 30));

        panelPValueRecalc.add(pvalueInfo);
        panelPValueRecalc.add(recalculate);
        springLayoutPanelPValueRecalc.putConstraint(SpringLayout.WEST,
                recalculate, 10,
                SpringLayout.EAST, pvalueInfo);
        springLayoutPanelPValueRecalc.putConstraint(SpringLayout.NORTH,
                recalculate, 0,
                SpringLayout.NORTH, pvalueInfo);

        SpringLayout springLayoutLabels = new SpringLayout();
        JPanel labels = new JPanel(springLayoutLabels);
        labels.setPreferredSize(new Dimension(400, 80));

        labels.add(panelPValueRecalc);
        labels.add(genesInPopulation);
        springLayoutLabels.putConstraint(SpringLayout.WEST, genesInPopulation,
                                         0,
                                         SpringLayout.WEST, panelPValueRecalc);
        springLayoutLabels.putConstraint(SpringLayout.NORTH, genesInPopulation,
                                         0,
                                         SpringLayout.SOUTH, panelPValueRecalc);

        labels.add(genesInBicluster);
        springLayoutLabels.putConstraint(SpringLayout.WEST, genesInBicluster, 0,
                                         SpringLayout.WEST, genesInPopulation);
        springLayoutLabels.putConstraint(SpringLayout.NORTH, genesInBicluster,
                                         10,
                                         SpringLayout.SOUTH, genesInPopulation);
        this.add(new JPanel());
        this.add(labels);

        this.paint(goProperties, go, bicluster.getGenesNames(),
                   pvalue);

        mainFrame.getTabbedPaneAnalyzing().setEnabledAt(2, true);
        mainFrame.getTabbedPaneAnalyzingFunction().setEnabledAt(0, true);
        mainFrame.getTabbedPaneAnalyzingFunction().setEnabledAt(1, false);
        return goProperties;
    }

    /**
     * Paints this panel with a table of faunctions for the given
     * <code>biclustering</code> (biclusters' group).
     *
     * @param biclustering the <code>Biclustering</code> (biclusters' group)
     * @param selectionPath the selection <code>TreePath</code>
     * @param biologicalRelevanceSummary the <code>Hashtable</code> which
     *               contains the actual table's info (a
     *               <code>BiologicalRelevanceSummary</code> for each
     *               bicluster in the biclusters' group; the hash key
     *               is the bicluster id)
     */
    public void paint(Biclustering biclustering, TreePath selectionPath,
                      Hashtable<Integer,BiologicalRelevanceSummary>
            biologicalRelevanceSummary) {
        this.nodeObject = biclustering;
        this.selectionPath = selectionPath;

        if (biologicalRelevanceSummary == null ||
            biologicalRelevanceSummary.isEmpty()) {
            this.paintFunctionAnalysisPanel();
            mainFrame.getTabbedPaneAnalyzingFunction().setEnabledAt(1, false);
            return;
        } else {
            JLabel pvalueInfo = new JLabel("Values calculated using p-value = "
                                           +
                                           biologicalRelevanceSummary.
                                           get(biologicalRelevanceSummary.
                                               keys().nextElement()).
                                           getSignificanceUserThreshold());
            JButton recalculate = new JButton("Recalculate");
            recalculate.addActionListener(
                    new
                    PanelAnalyzingFunctionTable_recalculate_actionAdapter(this));
            pvalueInfo.setPreferredSize(new Dimension(
                    (int) pvalueInfo.getPreferredSize().getWidth(),
                    (int) recalculate.getPreferredSize().getHeight()));

            SpringLayout springLayoutPanelPValueRecalc = new SpringLayout();
            JPanel panelPValueRecalc = new JPanel(springLayoutPanelPValueRecalc);
            panelPValueRecalc.setPreferredSize(new Dimension(200, 40));

            panelPValueRecalc.add(pvalueInfo);
            panelPValueRecalc.add(recalculate);
            springLayoutPanelPValueRecalc.putConstraint(SpringLayout.WEST,
                    recalculate, 10,
                    SpringLayout.EAST, pvalueInfo);
            springLayoutPanelPValueRecalc.putConstraint(SpringLayout.NORTH,
                    recalculate, 0,
                    SpringLayout.NORTH, pvalueInfo);

            try {
                TableFunctionGroup functionGroupTable = null;
                if (((NodeInfo) ((DefaultMutableTreeNode) selectionPath.
                                 getLastPathComponent()).getUserObject()).
                    getNodeObjectType() == NodeInfo.POSTPROCESSED_GROUP) {
                    functionGroupTable =
                            new
                            TableFunctionGroup(biclustering,
                                               biologicalRelevanceSummary, true);
                } else {
                    functionGroupTable =
                            new
                            TableFunctionGroup(biclustering,
                                               biologicalRelevanceSummary, false);
                }
                functionGroupTable.addMouseListener(
            new
            PanelAnalyzingFunctionTable_functionGroupTable_mouseAdapter(this));
                JScrollPane scrollTable = new JScrollPane(functionGroupTable);
                scrollTable.setBorder(BorderFactory.createEmptyBorder(20, 0, 20,
                        0));
                scrollTable.getVerticalScrollBar().
                setUnitIncrement(mainFrame.SCROLL_BAR_INCREMENT/2);

                this.add(new JPanel());
                this.add(new JPanel());
                this.add(panelPValueRecalc);
                this.add(scrollTable);
            } catch (InvalidNodeObjectTypeException ex) {
                ex.printStackTrace();
                return;
            }
        }
        mainFrame.getTabbedPaneAnalyzing().setEnabledAt(2, true);
        mainFrame.getTabbedPaneAnalyzingFunction().setEnabledAt(0, true);
        mainFrame.getTabbedPaneAnalyzingFunction().setEnabledAt(1, false);
    }

    /**
     * Puts a new GO frontend in the GO frontends hashtable.
     *
     * @param selectionPath the selection <code>TreePath</code>
     * @param go the <code>GOFrontEnd</code> for the given
     *           <code>selectionPath/code>
     */
    public void putGOFrontEnd(TreePath selectionPath, GOFrontEnd go) {
        String hashKey = this.getGOFrontEndHashKeyForSelectionPath(
                selectionPath);
        this.goFrontEnds.put(hashKey, go);
    }

    /**
     * Returns the hashtable key for the <code>GOFrontEnd</code> of
     * the given <code>selectionPath</code>.
     *
     * @param selectionPath the selection <code>TreePath</code>
     * @return String the hashtable's GO frontend key
     */
    public String getGOFrontEndHashKeyForSelectionPath(TreePath selectionPath) {
        NodeInfo node = (NodeInfo) ((DefaultMutableTreeNode) selectionPath.
                                    getLastPathComponent()).getUserObject();
        String hashKey;
        if (node.getNodeObjectType() == NodeInfo.BICLUSTER_SINGLE ||
            node.getNodeObjectType() == NodeInfo.POSTPROCESSED_SINGLE) {
            hashKey = selectionPath.getPath()[1].toString() + "_" +
                      node.getParentNodeInfo().toString();
        }
        else {
            hashKey = selectionPath.getPath()[1].toString() + "_" +
                      node.toString();
        }
        return hashKey;
    }

    /**
     * Returns the <code>GOFrontEnd</code> for the given
     * <code>selectionPath</code>, if it exists, otherwise returns null.
     *
     * @param selectionPath TreePath
     * @return GOFrontEnd
     */
    public GOFrontEnd getGOFrontEndForSelectionPath(TreePath selectionPath) {
        return goFrontEnds.get(this.getGOFrontEndHashKeyForSelectionPath(
                selectionPath));
    }

    /**
     * Calculates the GO frontend for the given bicluster.
     *
     * @param bicluster the <code>Bicluster</code>
     * @param pathTo_GeneOntologyFile <code>String</code> the path to gene
     *              ontology file
     * @param pathTo_GeneAssociationFile <code>String</code> the path to
     *              organism's specific gene association file
     * @return GOFrontEnd the GO frontend for the given bicluster
     */
    public GOFrontEnd calculateGOFrontEnd(Bicluster bicluster,
                                          String pathTo_GeneOntologyFile,
                                          String pathTo_GeneAssociationFile) {
        try {
            return new GOFrontEnd(((IMatrix) bicluster.
                                   getBiclusterSource().
                                   getMatrix()).getGenesNames(),
                                  pathTo_GeneOntologyFile,
                                  pathTo_GeneAssociationFile);
        }
        catch (IOException ex) {
            return null;
        }
        catch (OBOParserException ex) {
            return null;
        }
    }

    /**
     * Calculates the GO frontend for the given bicluster.
     *
     * @param bicluster the <code>Bicluster</code>
     * @return GOFrontEnd the GO frontend for the given bicluster
     */
    public GOFrontEnd calculateGOFrontEnd(Bicluster bicluster) {
        IMatrix matrix = (IMatrix) bicluster.getBiclusterSource().getMatrix();
        String organism = "";
        if (matrix instanceof AbstractExpressionMatrix) {
            organism = ((AbstractExpressionMatrix) matrix).getOrganism();
        } else {
            organism = ((DiscretizedExpressionMatrix) matrix).
                       getOriginalExpressionMatrix().getOrganism();
        }
        File geneOntologyFile = new File(mainFrame.getPathToGO_Files() +
                                         File.separator +
                                         PanelFunctionAnalysis.OBO_FILE_NAME);

        File geneAssociationFile = new File(mainFrame.getPathToGO_Files() +
                                            File.separator +
                                            "gene_association" +
                                            PanelFunctionAnalysis.
                                            findAnnotationFileExtensionForOrganism(
                organism));
        if (!geneOntologyFile.exists() || !geneAssociationFile.exists()) {
            return null;
        }
        return calculateGOFrontEnd(bicluster,
                                            geneOntologyFile.getAbsolutePath(),
                                            geneAssociationFile.getAbsolutePath());
    }

    /**
     * Calculates the GO frontend for the given biclusters' group.
     *
     * @param matrix the expression <code>IMatrix</code>
     * @param pathTo_GeneOntologyFile <code>String</code> the path to gene
     *              ontology file
     * @param pathTo_GeneAssociationFile <code>String</code> the path to
     *              organism's specific gene association file
     * @return GOFrontEnd the GO frontend for the given expression matrix
     */
    public GOFrontEnd calculateGOFrontEnd(IMatrix matrix,
                                          String pathTo_GeneOntologyFile,
                                          String pathTo_GeneAssociationFile) {
        try {
            return new GOFrontEnd(matrix.getGenesNames(),
                                  pathTo_GeneOntologyFile,
                                  pathTo_GeneAssociationFile);
        }
        catch (IOException ex) {
           return null;
        }
        catch (OBOParserException ex) {
            return null;
        }
    }

    /**
     * Calculates the GO frontend for the given expression matrix.
     *
     * @param matrix the expression <code>IMatrix</code>
     * @return GOFrontEnd the GO frontend for the given biclusters' group
     */
    public GOFrontEnd calculateGOFrontEnd(IMatrix matrix) {
        String organism = "";
        if (matrix instanceof AbstractExpressionMatrix) {
            organism = ((AbstractExpressionMatrix) matrix).getOrganism();
        } else {
            organism = ((DiscretizedExpressionMatrix) matrix).
                       getOriginalExpressionMatrix().getOrganism();
        }
        File geneOntologyFile = new File(mainFrame.getPathToGO_Files() +
                                         File.separator +
                                         PanelFunctionAnalysis.OBO_FILE_NAME);
        File geneAssociationFile = new File(mainFrame.getPathToGO_Files() +
                                            File.separator +
                                            "gene_association" +
                                            PanelFunctionAnalysis.
                                            findAnnotationFileExtensionForOrganism(
                organism));
        if (!geneOntologyFile.exists() || !geneAssociationFile.exists()) {
            return null;
        }

        return calculateGOFrontEnd(matrix,
                                   geneOntologyFile.getAbsolutePath(),
                                   geneAssociationFile.getAbsolutePath());
    }

    /**
     * Calculates the GO frontend for the given biclusters' group.
     *
     * @param biclustering <code>Biclustering</code> (the biclusters' group)
     * @param pathTo_GeneOntologyFile <code>String</code> the path to gene
     *              ontology file
     * @param pathTo_GeneAssociationFile <code>String</code> the path to
     *              organism's specific gene association file
     * @return GOFrontEnd the GO frontend for the given biclusters' group
     */
    public GOFrontEnd calculateGOFrontEnd(Biclustering biclustering,
                                          String pathTo_GeneOntologyFile,
                                          String pathTo_GeneAssociationFile) {
        try {
            return new GOFrontEnd(((IMatrix) biclustering.
                                   getMatrix()).getGenesNames(),
                                  pathTo_GeneOntologyFile,
                                  pathTo_GeneAssociationFile);
        }
        catch(IOException ex) {
            return null;
        }
        catch (OBOParserException ex) {
            return null;
        }
    }

    /**
     * Calculates the GO frontend for the given biclusters' group.
     *
     * @param biclustering <code>Biclustering</code> (the biclusters' group)
     * @return GOFrontEnd the GO frontend for the given biclusters' group
     */
    public GOFrontEnd calculateGOFrontEnd(Biclustering biclustering) {
        IMatrix matrix = (IMatrix) biclustering.getMatrix();
        String organism = "";
        if (matrix instanceof AbstractExpressionMatrix) {
            organism = ((AbstractExpressionMatrix) matrix).getOrganism();
        } else {
            organism = ((DiscretizedExpressionMatrix) matrix).
                       getOriginalExpressionMatrix().getOrganism();
        }
        File geneOntologyFile = new File(mainFrame.getPathToGO_Files() +
                                         File.separator +
                                         PanelFunctionAnalysis.OBO_FILE_NAME);
        File geneAssociationFile = new File(mainFrame.getPathToGO_Files() +
                                            File.separator +
                                            "gene_association" +
                                            PanelFunctionAnalysis.
                                            findAnnotationFileExtensionForOrganism(
                organism));
        if (!geneOntologyFile.exists() || !geneAssociationFile.exists()) {
            return null;
        }
        return calculateGOFrontEnd(biclustering,
                                   geneOntologyFile.getAbsolutePath(),
                                   geneAssociationFile.getAbsolutePath());
    }


    /**
     * Handles the mouse event fired when the user clicks the apply button.
     *
     * @param e the <code>ActionEvent</code> fired when the user clicks
     *          the apply button
     */
    public void applyButton_actionPerformed(ActionEvent e) {
        float pvalue = this.functionAnalysisPanel.checkPValue();
        if (pvalue < 0) {
            return;
        }

        GOFrontEnd go = this.functionAnalysisPanel.useGOFrontEnd();
        if (go != null) {
            applyFunctionAnalysis(go);
            return;
        }

        File[] goFiles = this.functionAnalysisPanel.checkGOFiles(false);
        if (goFiles == null) {
            return;
        }
        if (goFiles.length == 0) {
            return;
        }

        this.applyFunctionAnalysis(goFiles);
    }

    /**
     * Applies function analysis with function analysis options.
     *
     * @param go the <code>GOFrontEnd</code>
     */
    private void applyFunctionAnalysis(GOFrontEnd go) {
        float pvalue = this.functionAnalysisPanel.checkPValue();
        if (pvalue < 0) {
            return;
        }

        try {
            if (nodeObject instanceof Biclustering) {
                Object[] path = selectionPath.getPath();
                String pathToBiclusteringFunctionResults =
                        mainFrame.getPathToFunctionResults() +
                        File.separator + path[1].toString();

                File directoryBiclusteringFunctionResults = new File(
                        pathToBiclusteringFunctionResults);
                if (!directoryBiclusteringFunctionResults.exists()) {
                    directoryBiclusteringFunctionResults.mkdir();
                }

                ThreadFunctionAnalysis functionAnalysisThread =
                        new ThreadFunctionAnalysis(mainFrame,
                        (Biclustering) nodeObject,
                        selectionPath,
                        go,
                        directoryBiclusteringFunctionResults,
                        pvalue);
            }
            else
            if (nodeObject instanceof Bicluster) {
                Object[] path = selectionPath.getPath();
                String pathToBiclusteringFunctionResults =
                        mainFrame.getPathToFunctionResults() +
                        File.separator + path[1].toString() +
                        File.separator +
                        ((DefaultMutableTreeNode) path[path.
                         length - 2]).getUserObject();

                File directoryBiclusteringFunctionResults = new File(
                        pathToBiclusteringFunctionResults);
                if (!directoryBiclusteringFunctionResults.exists()) {
                    directoryBiclusteringFunctionResults.mkdir();
                }

                ThreadFunctionAnalysis functionAnalysisThread =
                        new ThreadFunctionAnalysis(mainFrame,
                        (Bicluster) nodeObject,
                        selectionPath,
                        go,
                        directoryBiclusteringFunctionResults,
                        pvalue);

            }
        } catch (Exception ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(mainFrame,
                                          "Unable to apply function analysis:\n" +
                                          ex.getMessage(),
                                          "Function analysis failed",
                                          JOptionPane.ERROR_MESSAGE);
        }
    }

    /**
     * Applies function analysis with function analysis options.
     *
     * @param goFiles File[] the gene ontology files<br/>
     *        goFiles[0] - gene ontology file<br/>
     *        goFiles[1] - gene association file
     */
    private void applyFunctionAnalysis(File[] goFiles) {
        float pvalue = this.functionAnalysisPanel.checkPValue();
        if (pvalue < 0) {
            return;
        }

        try {
            if (nodeObject instanceof Biclustering) {
                Object[] path = selectionPath.getPath();
                String pathToBiclusteringFunctionResults =
                        mainFrame.getPathToFunctionResults() +
                        File.separator + path[1].toString() /*+
                        File.separator + ((DefaultMutableTreeNode) selectionPath.
                                          getLastPathComponent())
                .getUserObject()*/;

                File directoryBiclusteringFunctionResults = new File(
                        pathToBiclusteringFunctionResults);
                if (!directoryBiclusteringFunctionResults.exists()) {
                    directoryBiclusteringFunctionResults.mkdir();
                }

                ThreadFunctionAnalysis functionAnalysisThread =
                        new ThreadFunctionAnalysis(mainFrame,
                        (Biclustering) nodeObject,
                        selectionPath,
                        goFiles,
                        directoryBiclusteringFunctionResults,
                        pvalue);
            }
            else
            if (nodeObject instanceof Bicluster) {
                Object[] path = selectionPath.getPath();
                String pathToBiclusteringFunctionResults =
                        mainFrame.getPathToFunctionResults() +
                        File.separator + path[1].toString() +
                        File.separator +
                        ((DefaultMutableTreeNode) path[path.
                         length - 2]).getUserObject();

                File directoryBiclusteringFunctionResults = new File(
                        pathToBiclusteringFunctionResults);
                if (!directoryBiclusteringFunctionResults.exists()) {
                    directoryBiclusteringFunctionResults.mkdir();
                }

                ThreadFunctionAnalysis functionAnalysisThread =
                        new ThreadFunctionAnalysis(mainFrame,
                        (Bicluster) nodeObject,
                        selectionPath,
                        goFiles,
                        directoryBiclusteringFunctionResults,
                        pvalue);

            }
        } catch (Exception ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(mainFrame,
                                          "Unable to apply function analysis:\n" +
                                          ex.getMessage(),
                                          "Function analysis failed",
                                          JOptionPane.ERROR_MESSAGE);
        }
    }

    /**
     * Repaints function analysis panel after recalculating
     * biological relevance of a bicluster. Mainly to be called by
     * <code>ThreadFunctionAnalysis</code>.
     *
     * @param goProperties the <code>GeneSetResult</code> for the bicluster
     * @param pvalue <code>float</code> the p-value user's threshold
     */
    public void repaintPanelAnalyzingFunctionTable(GeneSetResult goProperties,
            float pvalue) {
        if (nodeObject == null || !(nodeObject instanceof Bicluster) ||
            selectionPath == null) {
            return;
        }

        this.removeAll();
        this.functionAnalysisPanel.clear();
        this.paint((Bicluster)nodeObject, goProperties, pvalue, selectionPath);

        mainFrame.setVisible(true);
    }

    /**
     * Repaints function analysis panel after recalculating
     * biological relevance of a biclusters' group. Mainly to be called by
     * <code>ThreadFunctionAnalysis</code>.
     *
     * @param biologicalRelevanceSummary Hashtable
     */
    public void repaintPanelAnalyzingFunctionTable(
            Hashtable<Integer,BiologicalRelevanceSummary>
            biologicalRelevanceSummary) {
        if (nodeObject == null || !(nodeObject instanceof Biclustering) ||
            selectionPath == null) {
            return;
        }

        this.removeAll();
        this.functionAnalysisPanel.clear();
        this.paint((Biclustering)nodeObject, selectionPath,
                   biologicalRelevanceSummary);

        mainFrame.setVisible(true);
    }

    /**
     * Repaints function analysis panel after recalculating
     * biological relevance of an expression matrix (original,
     * preprocessed or discretized one). Mainly to be called by
     * <code>ThreadFunctionAnalysis</code>.
     */
    public void repaintPanelAnalyzingFunctionTable() {
        if (nodeObject == null ||
            !(nodeObject instanceof IMatrix) ||
            selectionPath == null) {
            return;
        }
        /**
         * paint the functions table again after recalculating functions
         */
        this.removeAll();
        this.functionAnalysisPanel.clear();
        this.paint((IMatrix) nodeObject, selectionPath);

        mainFrame.setVisible(true);
    }

    /**
     * Handles the mouse event fired when the user clicks on
     * recalculate functions button.
     *
     * @param e the <code>ActionEvent</code> fired when the user clicks on
     *          recalculate functions button
     */
    public void recalculate_actionPerformed(ActionEvent e) {
        this.removeAll();
        this.functionAnalysisPanel.clear();
        this.paintFunctionAnalysisPanel();
        mainFrame.setVisible(true);
    }

    /**
     * Handles the event fired when the mouse is clicked on a function
     * group table's row.
     * Each row corresponds to a bicluster. When the user clicks a given
     * row, the application redirects him/her to the corresponding bicluster
     * object.
     *
     * @param e the <code>MouseEvent</code> fired with the mouse's click
     */
    public void functionGroupTable_mouseClicked(MouseEvent e) {
        Point p = e.getPoint();
        TableFunctionGroup table = (TableFunctionGroup)e.getComponent();
        int viewRowIndex = table.rowAtPoint(p);
        int columnIndex = 0;
        if (((NodeInfo)((DefaultMutableTreeNode)selectionPath.
                        getLastPathComponent()).getUserObject()).
            getNodeObjectType() == NodeInfo.POSTPROCESSED_GROUP) {
            columnIndex++;
        }

        int nodeID = Integer.valueOf((Integer)table.getValueAt(viewRowIndex, columnIndex));
        Enumeration children = ((DefaultMutableTreeNode)selectionPath.getLastPathComponent()).children();

        int i = 0;
        while (children.hasMoreElements()) {
            DefaultMutableTreeNode node = (DefaultMutableTreeNode)children.nextElement();
            if (((NodeInfo)node.getUserObject()).getId() == nodeID) {
                Object[] oldPath = selectionPath.getPath();
                Object[] newPath = new Object[oldPath.length+1];
                int j;
                for (j = 0; j < oldPath.length; j++) {
                    newPath[j] = oldPath[j];
                }
                newPath[j] = node;
                TreePath path = new TreePath(newPath);
                mainFrame.getTree().setSelectionPath(path);
                break;
            }
            i++;
        }
    }
}


class PanelAnalyzingFunctionTable_recalculate_actionAdapter implements ActionListener {
    private PanelAnalyzingFunctionTable adaptee;
    PanelAnalyzingFunctionTable_recalculate_actionAdapter(
            PanelAnalyzingFunctionTable adaptee) {
        this.adaptee = adaptee;
    }

    public void actionPerformed(ActionEvent e) {
        adaptee.recalculate_actionPerformed(e);
    }
}


class PanelAnalyzingFunctionTable_applyButton_actionAdapter implements ActionListener {
    private PanelAnalyzingFunctionTable adaptee;
    PanelAnalyzingFunctionTable_applyButton_actionAdapter(
            PanelAnalyzingFunctionTable adaptee) {
        this.adaptee = adaptee;
    }

    public void actionPerformed(ActionEvent e) {
        adaptee.applyButton_actionPerformed(e);
    }
}

class PanelAnalyzingFunctionTable_functionGroupTable_mouseAdapter extends MouseAdapter {
    private PanelAnalyzingFunctionTable adaptee;
    PanelAnalyzingFunctionTable_functionGroupTable_mouseAdapter(
            PanelAnalyzingFunctionTable adaptee) {
        this.adaptee = adaptee;
    }

    /**
     *
     * @param e MouseEvent
     */
    public void mouseClicked(MouseEvent e) {
        adaptee.functionGroupTable_mouseClicked(e);
    }
}
