package biggests.tables;

import java.awt.*;
import java.awt.event.*;

import javax.swing.table.*;

import biggests.utils.*;
import smadeira.biclustering.*;
import smadeira.ontologizer.*;

/**
 * <p>Title: Table of values with missing values</p>
 *
 * <p>Description:
 *                 <p>Defines a table of values with the following data:<br/>
 *                 - header -> conditions' names (more widely, <code>String</code>s)<br/>
 *                 - first column -> genes' names (more widely, <code>String</code>s)<br/>
 *                 - all other columns' cells -> gene expression values
 *                 (more widely, <code>float</code> values)</p>
 *
 *                 <p>This is a simple table that shows the raw values that
 *                 are assigned when using one of the constructors and
 *                 implements sorting functionality on the columns, i.e.,
 *                 allows the reordering of the rows, sorting them by the
 *                 values of a specific column in ascending or descending
 *                 order.</p>
 *
 *                 <p>Specific for data with missing values, that is, some
 *                 erroneus values that are not part of the original data
 *                 and are marked by some float value.</p>
 * </p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonalves
 */
public class TableValuesWithMissingValues extends TableValues {
    /**
     * The value used to fill missing values.
     */
    float missingValue;

    /**
     * Creates a new values table from the data of an expression matrix with
     * missing values. A table of values contains the following data:<br/>
     * - header -> conditions' names<br/>
     * - first column -> genes' names<br/>
     * - all other column's cells -> gene expression float values<br/>
     * This table also displays the terms annotated with each gene
     * (when a specific table's row is clicked by the user).
     *
     * @param expressionMatrix the <code>ExpressionMatrix</code> with missing
     *                         values from which the table of values is created
     * @param go the <code>GOFrontEnd</code>
     * @param parent the GUI's parent <code>Component</code>
     * @throws InvalidNodeObjectTypeException when the type of the node
     *                   object is not valid (normally this is not a problem,
     *                   since an object of type <code>ExpressionMatrix</code>
     *                   is considered to be a valid type for a node object
     * @throws Exception
     * @see #TableValuesWithMissingValues(Object[][] rowData, Object[] columnNames))
     * @see TableValues#constructRowData(NodeObjectInterface nodeObject)
     * @see TableValues#constructColumnNames(NodeObjectInterface nodeObject)
     */
    /* ---------------------------------------------------------------------<br/>
     * | Genes Names | Condition1 | Condition2 | Condition3 | .............|<br/>
     * |-------------------------------------------------------------------|<br/>
     * | Gene1       | 0.26394536 | .......... | .......... | ............ |<br/>
     * | Gene2       | -0.0107532 | .......... | .......... | ............ |<br/>
     * | Gene3       | .......... | .......... | .......... | ............ |<br/>
     * | Gene4       | .......... | .......... | .......... | ............ |<br/>
     * ---------------------------------------------------------------------
     */
    public TableValuesWithMissingValues(IMatrix expressionMatrix,
                                        GOFrontEnd go, Component parent) throws
            InvalidNodeObjectTypeException, Exception {
        super(constructRowData(expressionMatrix),
              constructColumnNames(expressionMatrix));
        this.goFrontEnd = go;
        if ((expressionMatrix instanceof AbstractExpressionMatrix &&
            !((AbstractExpressionMatrix)expressionMatrix).hasMissingValues()) ||
            (expressionMatrix instanceof DiscretizedExpressionMatrix &&
             !((DiscretizedExpressionMatrix)expressionMatrix).
             getOriginalExpressionMatrix().hasMissingValues())) {
            throw (new Exception(
                    "Only expression matrices with missing values are allowed."));
        }
        if (expressionMatrix instanceof AbstractExpressionMatrix) {
            missingValue = ((AbstractExpressionMatrix)expressionMatrix).
                           getMissingValue();
        }
        else {
            missingValue = ((DiscretizedExpressionMatrix)expressionMatrix).
                           getOriginalExpressionMatrix().getMissingValue();
        }
        this.parent = parent;
    }

    /**
     * Creates a new values table from the data of an expression matrix with
     * missing values. A table of values contains the following data:<br/>
     * - header -> conditions' names<br/>
     * - first column -> genes' names<br/>
     * - all other column's cells -> gene expression float values<br/>
     *
     * @param expressionMatrix the <code>ExpressionMatrix</code> with missing
     *                         values from which the table of values is created
     * @throws InvalidNodeObjectTypeException when the type of the node
     *                   object is not valid (normally this is not a problem,
     *                   since an object of type <code>ExpressionMatrix</code>
     *                   is considered to be a valid type for a node object
     * @throws Exception
     * @see #TableValuesWithMissingValues(Object[][] rowData, Object[] columnNames))
     * @see TableValues#constructRowData(NodeObjectInterface nodeObject)
     * @see TableValues#constructColumnNames(NodeObjectInterface nodeObject)
     */
    /* ---------------------------------------------------------------------<br/>
     * | Genes Names | Condition1 | Condition2 | Condition3 | .............|<br/>
     * |-------------------------------------------------------------------|<br/>
     * | Gene1       | 0.26394536 | .......... | .......... | ............ |<br/>
     * | Gene2       | -0.0107532 | .......... | .......... | ............ |<br/>
     * | Gene3       | .......... | .......... | .......... | ............ |<br/>
     * | Gene4       | .......... | .......... | .......... | ............ |<br/>
     * ---------------------------------------------------------------------
     */
    public TableValuesWithMissingValues(IMatrix expressionMatrix) throws
            InvalidNodeObjectTypeException, Exception {
        super(constructRowData(expressionMatrix),
              constructColumnNames(expressionMatrix));
        if ((expressionMatrix instanceof AbstractExpressionMatrix &&
            !((AbstractExpressionMatrix)expressionMatrix).hasMissingValues()) ||
            (expressionMatrix instanceof DiscretizedExpressionMatrix &&
             !((DiscretizedExpressionMatrix)expressionMatrix).
             getOriginalExpressionMatrix().hasMissingValues())) {
            throw (new Exception(
                    "Only expression matrices with missing values are allowed."));
        }
        if (expressionMatrix instanceof AbstractExpressionMatrix) {
            missingValue = ((AbstractExpressionMatrix)expressionMatrix).
                           getMissingValue();
        }
        else {
            missingValue = ((DiscretizedExpressionMatrix)expressionMatrix).
                           getOriginalExpressionMatrix().getMissingValue();
        }
    }

    /**
     * Creates a new values table from the data of a bicluster with
     * missing values. A table of values contains the following data:<br/>
     * - header -> conditions' names<br/>
     * - first column -> genes' names<br/>
     * - all other column's cells -> gene expression float values<br/>
     * This table also displays the terms annotated with each gene
     * (when a specific table's row is clicked by the user).
     *
     * @param bicluster the <code>Bicluster</code> with missing
     *                         values from which the table of values is created
     * @param go the <code>GOFrontEnd</code>
     * @param parent the GUI's parent <code>Component</code>
     * @throws InvalidNodeObjectTypeException when the type of the node
     *                   object is not valid (normally this is not a problem,
     *                   since an object of type <code>ExpressionMatrix</code>
     *                   is considered to be a valid type for a node object
     * @throws Exception
     * @see #TableValuesWithMissingValues(Object[][] rowData, Object[] columnNames))
     * @see TableValues#constructRowData(NodeObjectInterface nodeObject)
     * @see TableValues#constructColumnNames(NodeObjectInterface nodeObject)
     */
    /* ---------------------------------------------------------------------<br/>
     * | Genes Names | Condition1 | Condition2 | Condition3 | .............|<br/>
     * |-------------------------------------------------------------------|<br/>
     * | Gene1       | 0.26394536 | .......... | .......... | ............ |<br/>
     * | Gene2       | -0.0107532 | .......... | .......... | ............ |<br/>
     * | Gene3       | .......... | .......... | .......... | ............ |<br/>
     * | Gene4       | .......... | .......... | .......... | ............ |<br/>
     * ---------------------------------------------------------------------
     */
    public TableValuesWithMissingValues(Bicluster bicluster,
                                        GOFrontEnd go, Component parent) throws
            InvalidNodeObjectTypeException, Exception {
        super(constructRowData(bicluster),
              constructColumnNames(bicluster));
        this.goFrontEnd = go;
        IMatrix matrix = (IMatrix)bicluster.getBiclusterSource().getMatrix();
        if ((matrix instanceof AbstractExpressionMatrix &&
            !((AbstractExpressionMatrix)matrix).hasMissingValues()) ||
            (matrix instanceof DiscretizedExpressionMatrix &&
             !((DiscretizedExpressionMatrix)matrix).
             getOriginalExpressionMatrix().hasMissingValues())) {
            throw (new Exception(
                    "Only expression matrices with missing values are allowed."));
        }
        if (matrix instanceof AbstractExpressionMatrix) {
            missingValue = ((AbstractExpressionMatrix)matrix).
                           getMissingValue();
        }
        else {
            missingValue = ((DiscretizedExpressionMatrix)matrix).
                           getOriginalExpressionMatrix().getMissingValue();
        }
        this.parent = parent;
    }

    /**
     * Creates a new values table from the data of a bicluster with
     * missing values. A table of values contains the following data:<br/>
     * - header -> conditions' names<br/>
     * - first column -> genes' names<br/>
     * - all other column's cells -> gene expression float values<br/>
     *
     * @param bicluster the <code>Bicluster</code> with missing
     *                         values from which the table of values is created
     * @throws InvalidNodeObjectTypeException when the type of the node
     *                   object is not valid (normally this is not a problem,
     *                   since an object of type <code>ExpressionMatrix</code>
     *                   is considered to be a valid type for a node object
     * @throws Exception
     * @see #TableValuesWithMissingValues(Object[][] rowData, Object[] columnNames))
     * @see TableValues#constructRowData(NodeObjectInterface nodeObject)
     * @see TableValues#constructColumnNames(NodeObjectInterface nodeObject)
     */
    /* ---------------------------------------------------------------------<br/>
     * | Genes Names | Condition1 | Condition2 | Condition3 | .............|<br/>
     * |-------------------------------------------------------------------|<br/>
     * | Gene1       | 0.26394536 | .......... | .......... | ............ |<br/>
     * | Gene2       | -0.0107532 | .......... | .......... | ............ |<br/>
     * | Gene3       | .......... | .......... | .......... | ............ |<br/>
     * | Gene4       | .......... | .......... | .......... | ............ |<br/>
     * ---------------------------------------------------------------------
     */
    public TableValuesWithMissingValues(Bicluster bicluster) throws
            InvalidNodeObjectTypeException, Exception {
        super(constructRowData(bicluster),
              constructColumnNames(bicluster));
        IMatrix matrix = (IMatrix)bicluster.getBiclusterSource().getMatrix();
        if ((matrix instanceof AbstractExpressionMatrix &&
            !((AbstractExpressionMatrix)matrix).hasMissingValues()) ||
            (matrix instanceof DiscretizedExpressionMatrix &&
             !((DiscretizedExpressionMatrix)matrix).
             getOriginalExpressionMatrix().hasMissingValues())) {
            throw (new Exception(
                    "Only expression matrices with missing values are allowed."));
        }
        if (matrix instanceof AbstractExpressionMatrix) {
            missingValue = ((AbstractExpressionMatrix)matrix).
                           getMissingValue();
        }
        else {
            missingValue = ((DiscretizedExpressionMatrix)matrix).
                           getOriginalExpressionMatrix().getMissingValue();
        }
    }

    /**
     * Overrides <code>TableValues</code>'s <code>prepareRenderer</code> method.
     * Prepares the renderer by querying the data model for the
     * value and selection state
     * of the cell at <code>row</code>, <code>column</code>.
     * Returns the component (may be a <code>Component</code>
     * or a <code>JComponent</code>) under the event location.
     *
     * @param renderer  the <code>TableCellRenderer</code> to prepare
     * @param row       the row of the cell to render, where 0 is the first row
     * @param column    the column of the cell to render,
     *			where 0 is the first column
     * @return          the <code>Component</code> under the event location
     * @see TableValues#prepareRenderer(TableCellRenderer renderer, int row, int column)
     */
    public Component prepareRenderer(TableCellRenderer renderer, int row,
                                     int column) {
        /**
         * get the component prepared from TableValues's prepareRenderer method
         */
        Component c = super.prepareRenderer(renderer, row, column);

        if (column > 0) {
            float value = ((Float)this.getValueAt(row, column)).floatValue();
            if (value == this.missingValue) {
                Color color = new Color(0, 0, 100);
                c.setBackground(color);
                c.setForeground(color);
            } else
            if ((row % 2 == 0) && (!isCellSelected(row, column))) {
                /**
                 * cell is in row numbered even and is not selected at the moment
                 * set the color of the cell to light blue
                 */
                c.setBackground(new Color(228, 238, 250));
                c.setForeground(getForeground());
            } else
            if (!isCellSelected(row, column)) {
                /**
                 * cell is in a row numbered odd and is not selected at the moment
                 * set the color of the cell to default background
                 */
                c.setBackground(getBackground());
                c.setForeground(getForeground());

            }
        }
        return c;
    }

    /**
     * Overrides <code>TableValues</code>'s <code>getToolTipText</code> method
     * in order to allow the renderer's tips to be used if they have text set.
     *
     * @param event MouseEvent the event that triggered the tool tip display
     * @return String the message set as tool tip for the pointed table cell
     * @see javax.swing.JTable#getToolTipText(MouseEvent event)
     */
    public String getToolTipText(MouseEvent event) {
        String tip = null;
        Point p = event.getPoint();
        int rowIndex = rowAtPoint(p);
        int colIndex = columnAtPoint(p);
        int realColumnIndex = convertColumnIndexToModel(colIndex);

        if (realColumnIndex == 0) {
            /**
             * cell is in the first column, so it contains the name of a gene
             * set the tex tip for this cell, something like: "Gene X"
             */
            tip = "Gene: " + getValueAt(rowIndex, 0);
        } else
        if ((rowIndex >= 0) && (rowIndex < this.getRowCount()) &&
            (realColumnIndex > 0) && (realColumnIndex < this.getColumnCount())) {
            /**
             * cell is in a column that is not the first one, so it contains
             * a float value (the gene expression value);
             * set the tip for this cell, something like:
             * "(gene_name, condition_name) = gene_expression_value"
             * if it's not a missing value, and:
             * "missing value"
             * it it is a missing value
             */
            if (((Float) getValueAt(rowIndex, realColumnIndex)).floatValue() ==
                missingValue) {
                tip = "missing value";
            } else {
                tip = "(" + getValueAt(rowIndex, 0) +
                      ", " + getColumnName(realColumnIndex) +
                      ") = " + getValueAt(rowIndex, realColumnIndex);
            }
        } else {
            tip = super.getToolTipText(event);
        }
        return tip;
    }
}
